import wrapt

from aws_xray_sdk.core import xray_recorder
from aws_xray_sdk.core.models import http
from aws_xray_sdk.ext.util import inject_trace_header, strip_url


def patch():

    wrapt.wrap_function_wrapper(
        'requests',
        'Session.request',
        _xray_traced_requests
    )

    wrapt.wrap_function_wrapper(
        'requests',
        'Session.prepare_request',
        _inject_header
    )


def _xray_traced_requests(wrapped, instance, args, kwargs):

    url = kwargs.get('url') or args[1]

    return xray_recorder.record_subsegment(
        wrapped, instance, args, kwargs,
        name=strip_url(url),
        namespace='remote',
        meta_processor=requests_processor,
    )


def _inject_header(wrapped, instance, args, kwargs):
    request = args[0]
    headers = getattr(request, 'headers', {})
    inject_trace_header(headers, xray_recorder.current_subsegment())
    setattr(request, 'headers', headers)

    return wrapped(*args, **kwargs)


def requests_processor(wrapped, instance, args, kwargs,
                       return_value, exception, subsegment, stack):

    method = kwargs.get('method') or args[0]
    url = kwargs.get('url') or args[1]

    subsegment.put_http_meta(http.METHOD, method)
    subsegment.put_http_meta(http.URL, url)

    if return_value is not None:
        subsegment.put_http_meta(http.STATUS, return_value.status_code)
    elif exception:
        subsegment.add_exception(exception, stack)



def wildcard_match(pattern, text, case_insensitive=True):
    """
    Performs a case-insensitive wildcard match against two strings.
    This method works with pseduo-regex chars; specifically ? and * are supported.
    An asterisk (*) represents any combination of characters.
    A question mark (?) represents any single character.
    :param str pattern: the regex-like pattern to be compared against
    :param str text: the string to compare against the pattern
    :param boolean case_insensitive: dafault is True
    return whether the text matches the pattern
    """
    if pattern is None or text is None:
        return False

    pattern_len = len(pattern)
    text_len = len(text)
    if pattern_len == 0:
        return text_len == 0

    # Check the special case of a single * pattern, as it's common
    if pattern == '*':
        return True

    if case_insensitive:
        pattern = pattern.lower()
        text = text.lower()

    # Infix globs are relatively rare, and the below search is expensive.
    # Check for infix globs and, in their absence, do the simple thing.
    if '*' not in pattern or pattern.index('*') == len(pattern) - 1:
        return _simple_wildcard_match(pattern, text)

    # The res[i] is used to record if there is a match between
    # the first i chars in text and the first j chars in pattern.
    # So will return res[textLength+1] in the end
    # Loop from the beginning of the pattern
    # case not '*': if text[i]==pattern[j] or pattern[j] is '?',
    # and res[i] is true, set res[i+1] to true, otherwise false.
    # case '*': since '*' can match any globing, as long as there is a true
    # in res before i, all the res[i+1], res[i+2],...,res[textLength]
    # could be true
    res = [None] * (text_len + 1)
    res[0] = True
    for j in range(0, pattern_len):
        p = pattern[j]
        if p != '*':
            for i in range(text_len - 1, -1, -1):
                res[i + 1] = res[i] and (p == '?' or (p == text[i]))
        else:
            i = 0
            while i <= text_len and not res[i]:
                i += 1
            for m in range(i, text_len + 1):
                res[m] = True

        res[0] = res[0] and (p == '*')

    return res[text_len]


def _simple_wildcard_match(pattern, text):
    j = 0
    pattern_len = len(pattern)
    text_len = len(text)
    for i in range(0, pattern_len):
        p = pattern[i]
        if p == '*':
            # Presumption for this method is that globs only occur at end
            return True
        elif p == '?':
            if j == text_len:
                # No character to match
                return False
            j += 1
        else:
            if j >= text_len:
                return False

            if(p != text[j]):
                return False
            j += 1

    # Ate up all the pattern and didn't end at a glob, so a match
    # will have consumed all the text
    return j == text_len



import binascii
import os
import time
from datetime import datetime

import botocore.session
from botocore import UNSIGNED
from botocore.client import Config

from .sampling_rule import SamplingRule
from aws_xray_sdk.core.models.dummy_entities import DummySegment
from aws_xray_sdk.core.context import Context
from aws_xray_sdk.core.utils.compat import PY2


class ServiceConnector(object):
    """
    Connector class that translates Centralized Sampling poller functions to
    actual X-Ray back-end APIs and communicates with X-Ray daemon as the
    signing proxy.
    """
    def __init__(self):
        self._xray_client = self._create_xray_client()
        self._client_id = binascii.b2a_hex(os.urandom(12)).decode('utf-8')
        self._context = Context()

    def _context_wrapped(func):
        """
        Wrapping boto calls with dummy segment. This is because botocore
        has two dependencies (requests and httplib) that might be
        monkey-patched in user code to capture subsegments. The wrapper
        makes sure there is always a non-sampled segment present when
        the connector makes an  AWS API call using botocore.
        This context wrapper doesn't work with asyncio based context
        as event loop is not thread-safe.
        """
        def wrapper(self, *args, **kargs):
            if type(self.context).__name__ == 'AsyncContext':
                return func(self, *args, **kargs)
            segment = DummySegment()
            self.context.set_trace_entity(segment)
            result = func(self, *args, **kargs)
            self.context.clear_trace_entities()
            return result

        return wrapper

    @_context_wrapped
    def fetch_sampling_rules(self):
        """
        Use X-Ray botocore client to get the centralized sampling rules
        from X-Ray service. The call is proxied and signed by X-Ray Daemon.
        """
        new_rules = []

        resp = self._xray_client.get_sampling_rules()
        records = resp['SamplingRuleRecords']

        for record in records:
            rule_def = record['SamplingRule']
            if self._is_rule_valid(rule_def):
                rule = SamplingRule(name=rule_def['RuleName'],
                                    priority=rule_def['Priority'],
                                    rate=rule_def['FixedRate'],
                                    reservoir_size=rule_def['ReservoirSize'],
                                    host=rule_def['Host'],
                                    service=rule_def['ServiceName'],
                                    method=rule_def['HTTPMethod'],
                                    path=rule_def['URLPath'],
                                    service_type=rule_def['ServiceType'])
                new_rules.append(rule)

        return new_rules

    @_context_wrapped
    def fetch_sampling_target(self, rules):
        """
        Report the current statistics of sampling rules and
        get back the new assgiend quota/TTL froom the X-Ray service.
        The call is proxied and signed via X-Ray Daemon.
        """
        now = int(time.time())
        report_docs = self._generate_reporting_docs(rules, now)
        resp = self._xray_client.get_sampling_targets(
            SamplingStatisticsDocuments=report_docs
        )
        new_docs = resp['SamplingTargetDocuments']

        targets_mapping = {}
        for doc in new_docs:
            TTL = self._dt_to_epoch(doc['ReservoirQuotaTTL']) if doc.get('ReservoirQuotaTTL', None) else None
            target = {
                'rate': doc['FixedRate'],
                'quota': doc.get('ReservoirQuota', None),
                'TTL': TTL,
                'interval': doc.get('Interval', None),
            }
            targets_mapping[doc['RuleName']] = target

        return targets_mapping, self._dt_to_epoch(resp['LastRuleModification'])

    def setup_xray_client(self, ip, port, client):
        """
        Setup the xray client based on ip and port.
        If a preset client is specified, ip and port
        will be ignored.
        """
        if not client:
            client = self._create_xray_client(ip, port)
        self._xray_client = client

    @property
    def context(self):
        return self._context

    @context.setter
    def context(self, v):
        self._context = v

    def _generate_reporting_docs(self, rules, now):
        report_docs = []

        for rule in rules:
            statistics = rule.snapshot_statistics()
            doc = {
                'RuleName': rule.name,
                'ClientID': self._client_id,
                'RequestCount': statistics['request_count'],
                'BorrowCount': statistics['borrow_count'],
                'SampledCount': statistics['sampled_count'],
                'Timestamp': now,
            }
            report_docs.append(doc)
        return report_docs

    def _dt_to_epoch(self, dt):
        """
        Convert a offset-aware datetime to POSIX time.
        """
        if PY2:
            # The input datetime is from botocore unmarshalling and it is
            # offset-aware so the timedelta of subtracting this time
            # to 01/01/1970 using the same tzinfo gives us
            # Unix Time (also known as POSIX Time).
            time_delta = dt - datetime(1970, 1, 1).replace(tzinfo=dt.tzinfo)
            return int(time_delta.total_seconds())
        else:
            # Added in python 3.3+ and directly returns POSIX time.
            return int(dt.timestamp())

    def _is_rule_valid(self, record):
        # We currently only handle v1 sampling rules.
        return record.get('Version', None) == 1 and \
            record.get('ResourceARN', None) == '*' and \
            record.get('ServiceType', None) and \
            not record.get('Attributes', None)

    def _create_xray_client(self, ip='127.0.0.1', port='2000'):
        session = botocore.session.get_session()
        url = 'http://%s:%s' % (ip, port)
        return session.create_client('xray', endpoint_url=url,
                                     region_name='us-west-2',
                                     config=Config(signature_version=UNSIGNED)
                                     )



from .segment import Segment
from ..exceptions.exceptions import FacadeSegmentMutationException


MUTATION_UNSUPPORTED_MESSAGE = 'FacadeSegments cannot be mutated.'


class FacadeSegment(Segment):
    """
    This type of segment should only be used in an AWS Lambda environment.
    It holds the same id, traceid and sampling decision as
    the segment generated by Lambda service but its properties cannot
    be mutated except for its subsegments. If this segment is created
    before Lambda worker finishes initializatioin, all the child
    subsegments will be discarded.
    """
    def __init__(self, name, entityid, traceid, sampled):

        self.initializing = self._is_initializing(
            entityid=entityid,
            traceid=traceid,
            sampled=sampled,
        )

        super(FacadeSegment, self).__init__(
            name=name,
            entityid=entityid,
            traceid=traceid,
            sampled=sampled,
        )

    def close(self, end_time=None):
        """
        Unsupported operation. Will raise an exception.
        """
        raise FacadeSegmentMutationException(MUTATION_UNSUPPORTED_MESSAGE)

    def put_http_meta(self, key, value):
        """
        Unsupported operation. Will raise an exception.
        """
        raise FacadeSegmentMutationException(MUTATION_UNSUPPORTED_MESSAGE)

    def put_annotation(self, key, value):
        """
        Unsupported operation. Will raise an exception.
        """
        raise FacadeSegmentMutationException(MUTATION_UNSUPPORTED_MESSAGE)

    def put_metadata(self, key, value, namespace='default'):
        """
        Unsupported operation. Will raise an exception.
        """
        raise FacadeSegmentMutationException(MUTATION_UNSUPPORTED_MESSAGE)

    def set_aws(self, aws_meta):
        """
        Unsupported operation. Will raise an exception.
        """
        raise FacadeSegmentMutationException(MUTATION_UNSUPPORTED_MESSAGE)

    def set_user(self, user):
        """
        Unsupported operation. Will raise an exception.
        """
        raise FacadeSegmentMutationException(MUTATION_UNSUPPORTED_MESSAGE)

    def add_throttle_flag(self):
        """
        Unsupported operation. Will raise an exception.
        """
        raise FacadeSegmentMutationException(MUTATION_UNSUPPORTED_MESSAGE)

    def add_fault_flag(self):
        """
        Unsupported operation. Will raise an exception.
        """
        raise FacadeSegmentMutationException(MUTATION_UNSUPPORTED_MESSAGE)

    def add_error_flag(self):
        """
        Unsupported operation. Will raise an exception.
        """
        raise FacadeSegmentMutationException(MUTATION_UNSUPPORTED_MESSAGE)

    def add_exception(self, exception, stack, remote=False):
        """
        Unsupported operation. Will raise an exception.
        """
        raise FacadeSegmentMutationException(MUTATION_UNSUPPORTED_MESSAGE)

    def apply_status_code(self, status_code):
        """
        Unsupported operation. Will raise an exception.
        """
        raise FacadeSegmentMutationException(MUTATION_UNSUPPORTED_MESSAGE)

    def serialize(self):
        """
        Unsupported operation. Will raise an exception.
        """
        raise FacadeSegmentMutationException(MUTATION_UNSUPPORTED_MESSAGE)

    def ready_to_send(self):
        """
        Facade segment should never be sent out. This always
        return False.
        """
        return False

    def increment(self):
        """
        Increment total subsegments counter by 1.
        """
        self._subsegments_counter.increment()

    def decrement_ref_counter(self):
        """
        No-op
        """
        pass

    def _is_initializing(self, entityid, traceid, sampled):
        return not entityid or not traceid or sampled is None



import pg8000
import wrapt

from aws_xray_sdk.ext.dbapi2 import XRayTracedConn
from aws_xray_sdk.core.patcher import _PATCHED_MODULES
from aws_xray_sdk.ext.util import unwrap


def patch():

    wrapt.wrap_function_wrapper(
        'pg8000',
        'connect',
        _xray_traced_connect
    )


def _xray_traced_connect(wrapped, instance, args, kwargs):

    conn = wrapped(*args, **kwargs)
    meta = {
        'database_type': 'PostgreSQL',
        'user': conn.user.decode('utf-8'),
        'driver_version': 'Pg8000'
    }

    if hasattr(conn, '_server_version'):
        version = getattr(conn, '_server_version')
        if version:
            meta['database_version'] = str(version)

    return XRayTracedConn(conn, meta)


def unpatch():
    """
    Unpatch any previously patched modules.
    This operation is idempotent.
    """
    _PATCHED_MODULES.discard('pg8000')
    unwrap(pg8000, 'connect')
